
import re
from collections import defaultdict
from math import isclose
from textwrap import dedent

from django.db import transaction
from django.db.models import Count
from drf_spectacular.utils import extend_schema_field
from rest_framework import permissions, serializers
from rest_framework.exceptions import ValidationError
from rest_framework.validators import UniqueTogetherValidator

from arkindex.documents.models import Corpus, Element
from arkindex.documents.serializers.elements import ElementListSerializer
from arkindex.ponos.models import Task
from arkindex.process.models import Worker
from arkindex.project.serializer_fields import ArchivedField, DatasetSetsCountField, EnumField
from arkindex.project.tools import add_as_prefetch
from arkindex.training.models import (
    Dataset,
    DatasetElement,
    DatasetSet,
    DatasetState,
    Model,
    ModelVersion,
    ModelVersionState,
)
from arkindex.users.models import Role
from arkindex.users.utils import get_max_level


def _model_from_context(serializer_field):
    if isinstance(serializer_field.parent.instance, ModelVersion):
        return serializer_field.parent.instance.model
    return serializer_field.context.get("model")


def _corpus_from_context(serializer_field):
    if isinstance(serializer_field.parent.instance, Dataset):
        return serializer_field.parent.instance.corpus
    return serializer_field.context.get("corpus")


def _dataset_from_context(serializer_field):
    return serializer_field.context.get("dataset")


_model_from_context.requires_context = True
_corpus_from_context.requires_context = True
_dataset_from_context.requires_context = True


class ModelLightSerializer(serializers.ModelSerializer):
    name = serializers.CharField(max_length=100)

    class Meta:
        model = Model
        fields = ("id", "name")

    def validate_name(self, name):
        """Ensure model's unique name is available"""
        # Name unicity check is performed at the view level for creation
        if not self.instance:
            return name
        if Model.objects.using("default").filter(name=name).exclude(id=self.instance.id).exists():
            raise ValidationError("A model with this name already exists")
        return name


class ModelSerializer(ModelLightSerializer):
    # Actually define the field to avoid the field-level automatically generated UniqueValidator
    rights = serializers.SerializerMethodField(read_only=True)

    archived = ArchivedField(
        required=False,
        help_text=dedent("""
            Whether this model is archived.
            Model versions cannot be created on archived models, and they cannot be used in processes.

            Archiving or unarchiving an existing model requires admin access to the model.
        """),
    )

    class Meta(ModelLightSerializer.Meta):
        fields = ModelLightSerializer.Meta.fields + ("created", "updated", "archived", "description", "rights")

    def create(self, validated_data):
        instance = super().create(validated_data)

        # Create admin membership for creator user
        instance.memberships.create(
            user=self.context["request"].user,
            level=Role.Admin.value
        )
        return instance

    @extend_schema_field(serializers.ListField(child=serializers.ChoiceField(["read", "write", "admin"])))
    def get_rights(self, model):
        # to avoid a stale read on get_max_level when creating a model (because the corresponding membership is
        # created at the same time as the model) if a model is being created we return these rights without
        # actually checking them.
        if self.context["request"].method == "POST":
            return ["read", "write", "admin"]

        level = get_max_level(self.context["request"].user, model)
        rights = ["read"]
        if level >= Role.Contributor.value:
            rights.append("write")
        if level >= Role.Admin.value:
            rights.append("admin")
        return rights


class ModelCreateSerializer(ModelSerializer):
    # Exclude archived from model creation, since creating an archived model makes no sense
    archived = ArchivedField(read_only=True)


class CreateModelErrorResponseSerializer(serializers.Serializer):
    id = serializers.UUIDField(required=False, help_text="UUID of an existing model, if the error comes from a duplicate name.")
    name = serializers.CharField(required=False, help_text="Name of an existing model, if the error comes from a duplicate name.")
    detail = serializers.CharField(required=False, help_text="A generic error message when an error occurs outside of a specific field.")


class ModelCompatibleWorkerSerializer(serializers.ModelSerializer):
    model = serializers.PrimaryKeyRelatedField(queryset=Model.objects.none())
    worker = serializers.PrimaryKeyRelatedField(queryset=Worker.objects.none())

    class Meta:
        model = Model.compatible_workers.through
        fields = ("model", "worker")
        # Disable the implicit UniqueTogetherValidator that DRF adds,
        # as it will also run for deletions and always fail
        validators = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.context.get("request"):
            user = self.context["request"].user
            # Editing the compatible workers is editing an attribute of a model
            # Avoid a stale read if we just created a model and and want to add compatible workers to it
            self.fields["model"].queryset = Model.objects.editable(user).using("default")
            # Any worker that the user can browse can be set
            self.fields["worker"].queryset = Worker.objects.executable(user)

    def validate_model(self, model):
        if model.archived:
            raise ValidationError("This model is archived.")
        return model

    def validate(self, data):
        data = super().validate(data)
        # Ignore the uniqueness check for deletions
        if self.context["request"].method == "DELETE":
            return data

        # Avoid a stale read if creation is called multiple times in a row
        if data["model"].compatible_workers.filter(id=data["worker"].id).using("default").exists():
            raise ValidationError({"worker": ["This worker is already marked as compatible with this model."]})

        return data


class ModelVersionLightSerializer(serializers.ModelSerializer):
    """
    Serialize a simple model version
    """

    model = ModelLightSerializer(default=_model_from_context)
    state = EnumField(ModelVersionState)
    configuration = serializers.DictField(style={"base_template": "textarea.html"})

    class Meta:
        model = ModelVersion
        fields = ("id", "model", "tag", "state", "size", "configuration")
        read_only_fields = ("id", "model", "state", "size", "hash", "archive_hash")


class ModelVersionSerializer(serializers.ModelSerializer):
    """
    Serialize a model version with fields that can be updated regardless of its state.
    """
    model = serializers.HiddenField(default=_model_from_context)
    parent = serializers.PrimaryKeyRelatedField(
        queryset=ModelVersion.objects.none(),
        default=None,
        allow_null=True,
    )
    description = serializers.CharField(required=False, style={"base_template": "textarea.html"})
    configuration = serializers.DictField(required=False, style={"base_template": "textarea.html"})
    tag = serializers.CharField(allow_null=True, max_length=50, required=False, default=None)
    state = EnumField(ModelVersionState, required=False)
    s3_url = serializers.SerializerMethodField(read_only=True, help_text="Only returned if the user has **contributor** access to the model, and the model is available.")
    s3_put_url = serializers.SerializerMethodField(read_only=True, help_text="Only returned if the user has **contributor** access to the model, and the model is **not** available.")
    hash = serializers.RegexField(re.compile(r"[0-9A-Fa-f]{32}"), min_length=32, max_length=32, required=False)
    archive_hash = serializers.RegexField(re.compile(r"[0-9A-Fa-f]{32}"), min_length=32, max_length=32, required=False, write_only=True)

    class Meta:
        model = ModelVersion
        fields = ("id", "model", "model_id", "parent", "description", "tag", "state", "size", "hash", "archive_hash", "configuration", "s3_url", "s3_put_url", "created")
        read_only_fields = ("id", "model")
        validators = [
            UniqueTogetherValidator(
                queryset=ModelVersion.objects.filter(tag__isnull=False),
                fields=["model", "tag"],
                message="A version for this model with this tag already exists."
            )
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if isinstance(self.instance, ModelVersion):
            model = self.instance.model
        else:
            model = self.context.get("model")
        if model:
            qs = ModelVersion.objects.filter(model__in=Model.objects.readable(self.context["request"].user))
            if getattr(self.instance, "id", None):
                qs = qs.exclude(id=self.instance.id)
            self.fields["parent"].queryset = qs

    def validate(self, data):
        model = data.get("model")
        if not model:
            model = self.instance.model

        # Raise right away if the parent model is archived
        if model.archived:
            raise ValidationError({"model": ["This model is archived."]})

        errors = defaultdict(list)

        # Check hash unicity
        version_hash = data.get("hash")
        if version_hash:
            existing_model_version = model.versions.using("default").filter(state=ModelVersionState.Available, hash=version_hash).first()
            if existing_model_version:
                # If creating a new model version, raise the exception
                if not self.instance:
                    raise ValidationError({
                        "hash": "A model version with this hash already exists.",
                        "id": existing_model_version.id
                    })
                errors["hash"].append("A model version with this hash already exists.")
                errors["id"] = existing_model_version.id
            else:
                if self.instance:
                    self.instance.hash = version_hash

        # The following checks are only done when updating an existing version
        if not self.instance:
            return data

        # Check that state update is allowed
        new_state = data.get("state")
        if (
            new_state
            and new_state != self.instance.state
            and (new_state != ModelVersionState.Available or self.instance.state != ModelVersionState.Created)
        ):
            errors["state"].append(
                f"You can only update a model version's state from Created to Available, not {self.instance.state} to {new_state}."
            )

        # Check file size and archive hash when a version is made Available,
        # or if the version already is available and either the size or archive_hash are being updated
        archive_hash = data.get("archive_hash")
        size = data.get("size")
        check_errors = 0
        if (
            self.instance.state == ModelVersionState.Available
            and not new_state
            and (archive_hash or size)
        ) or new_state == ModelVersionState.Available:

            # Update the instance with the new archive_hash and size, if sent, so they get checked
            if archive_hash:
                self.instance.archive_hash = archive_hash
            if size:
                self.instance.size = size

            try:
                self.instance.perform_check()
            except AssertionError as e:
                check_errors += 1
                errors["size"].append(e)
            try:
                self.instance.check_hash()
            except (AssertionError, ValueError) as e:
                check_errors += 1
                errors["archive_hash"].append(e)

        if check_errors > 0:
            # Set the model version as erroneous
            self.instance.state = ModelVersionState.Error
            self.instance.save(update_fields=["state"])

        if errors:
            raise ValidationError(errors)

        return data

    @extend_schema_field(serializers.CharField(allow_null=True))
    def get_s3_put_url(self, obj):
        if not self.context.get("is_contributor") or obj.state == ModelVersionState.Available:
            return None
        return obj.s3_put_url

    @extend_schema_field(serializers.CharField(allow_null=True))
    def get_s3_url(self, obj):
        if not self.context.get("is_contributor") or obj.state != ModelVersionState.Available:
            return None
        return obj.s3_url


class DatasetSetSerializer(serializers.ModelSerializer):
    dataset = serializers.HiddenField(
        default=_dataset_from_context,
        write_only=True,
    )

    class Meta:
        model = DatasetSet
        fields = ("id", "name", "dataset", )
        read_only_fields = ("id", )
        # Remove the DRF-provided unique validator, which causes extra queries.
        # We do our own unique validation using the already prefetched sets on the dataset.
        validators = []

    def validate_name(self, name):
        if any(ds.name == name for ds in self.context["dataset"].sets.all()):
            raise ValidationError("A set with this name already exists in this dataset.")
        return name


class DatasetSerializer(serializers.ModelSerializer):
    state = EnumField(
        DatasetState,
        # Spectacular does not understand CreateOnlyDefault, so we document it in the help text too
        default=serializers.CreateOnlyDefault(DatasetState.Open),
        help_text=dedent("""
        When using Ponos task authentication, state update is limited to these transitions:

            Open → Building → Complete
                      ↕
                    Error

        When not using Ponos task authentication, state can be updated following those transitions:

            Complete
                   ↘
            Error → Open

        When updated to `complete` through Ponos task authentication, the dataset is automatically
        linked to the task used for authentication. In any other case, it is set to None.

        When creating, this is always set to `open`.
        """),
    )

    creator = serializers.CharField(
        source="creator.display_name",
        read_only=True,
        default=serializers.CurrentUserDefault(),
        help_text="Display name of the user who created the dataset.",
    )

    set_names = serializers.ListField(
        child=serializers.CharField(max_length=50),
        write_only=True,
        default=serializers.CreateOnlyDefault(["train", "dev", "test"])
    )
    sets = DatasetSetSerializer(many=True, read_only=True)

    # When creating the dataset, the dataset's corpus comes from the URL, so the APIView passes it through
    corpus = serializers.HiddenField(default=_corpus_from_context)

    # When creating the dataset, the dataset's creator is set to the current user automatically
    default_creator = serializers.HiddenField(
        source="creator",
        default=serializers.CreateOnlyDefault(serializers.CurrentUserDefault()),
    )
    set_elements = DatasetSetsCountField(
        help_text="Distribution of elements in sets. This value is set to null when listing multiple datasets.",
    )
    unique_elements = serializers.BooleanField(
        default=True,
        help_text="Ensures that an element is only present in a single set at a time.",
    )

    def validate_state(self, state):
        # Ensure that only the open state can be set when creating
        if not isinstance(self.instance, Dataset):
            return DatasetState.Open

        errors = []

        if state == self.instance.state:
            return state

        request = self.context.get("request")
        if isinstance(request.auth, Task):
            transitions = {
                DatasetState.Open: (DatasetState.Building,),
                DatasetState.Building: (DatasetState.Complete, DatasetState.Error),
                DatasetState.Error: (DatasetState.Building,),
            }
            if not request.auth.process.sets.filter(dataset_id=self.instance.id).exists():
                errors.append("A task can only update the state of one of the datasets of its process.")
        else:
            transitions = {
                DatasetState.Error: (DatasetState.Open,),
                DatasetState.Complete: (DatasetState.Open,),
            }

        if state not in transitions.get(self.instance.state, ()):
            errors.append(f"Transition from {self.instance.state} to {state} is not allowed.")

        if errors:
            raise ValidationError(errors)
        return state

    def validate_set_names(self, set_names):
        if self.context["request"].method not in permissions.SAFE_METHODS and self.context["request"].method != "POST":
            raise ValidationError("This API endpoint does not allow updating a dataset's sets.")
        if set_names is not None and len(set(set_names)) != len(set_names):
            raise ValidationError("Set names must be unique.")
        if set_names is not None and len(set_names) == 0:
            raise ValidationError("Either do not specify set names to use the default values, or specify a non-empty list of names.")
        return set_names

    def validate_unique_elements(self, unique):
        # When updating a dataset to switch unique_elements from False to True,
        # check that it does not contain duplicates.
        if unique is True and self.instance and not self.instance.unique_elements and (
            DatasetElement.objects
            .filter(set__dataset_id=self.instance.pk)
            .values("element_id")
            .annotate(dupes=Count("element_id"))
            .filter(dupes__gte=2)
            .exists()
        ):
            raise ValidationError("Some elements are currently contained by multiple sets.")
        return unique

    def validate(self, data):
        data = super().validate(data)

        # No further validation needed when creating
        if not isinstance(self.instance, Dataset):
            return data

        state = data.get("state")
        request = self.context.get("request")
        if isinstance(request.auth, Task) and state == DatasetState.Complete:
            # Link a completed dataset to the current task which generated its artifacts
            data["task_id"] = request.auth.id
        elif state == DatasetState.Open:
            # Delete eventual linked artifacts
            data["task_id"] = None

        return data

    @transaction.atomic
    def create(self, validated_data):
        set_names = validated_data.pop("set_names")
        dataset = Dataset.objects.create(**validated_data)
        sets = DatasetSet.objects.bulk_create(
            DatasetSet(
                name=set_name,
                dataset_id=dataset.id
            ) for set_name in sorted(set_names)
        )
        # We will output set element counts in the API, but we know there are zero,
        # so no need to make another query to prefetch the sets and count them
        for set in sets:
            set.element_count = 0
        add_as_prefetch(dataset.sets, sets)
        return dataset

    class Meta:
        model = Dataset
        fields = (
            "id",
            "name",
            "description",
            "sets",
            "set_names",
            "set_elements",
            "state",
            # Only the corpus ID is actually serialized
            "corpus_id",
            # Hidden field retrieved from the API's context
            "corpus",
            "creator",
            # Hidden field to set the creator as the authenticated user
            "default_creator",
            "task_id",
            "unique_elements",
            "created",
            "updated",
        )
        read_only_fields = (
            "id",
            "corpus_id",
            "corpus",
            "creator",
            "default_creator",
            "task_id",
            "created",
            "updated",
        )
        validators = [
            UniqueTogetherValidator(
                queryset=Dataset.objects.all(),
                fields=["corpus", "name"],
                message="A dataset with this name already exists in this corpus."
            )
        ]
        extra_kwargs = {
            "sets": {
                "error_messages": {
                    "empty": "Either do not specify set names to use the default values, or specify a non-empty list of names."
                }
            }
        }


class DatasetElementSerializer(serializers.ModelSerializer):
    element = ElementListSerializer(
        read_only=True,
        allow_null=False,
    )
    element_id = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.none(),
        style={"base_template": "input.html"},
        source="element",
        write_only=True,
    )
    dataset = serializers.HiddenField(
        default=_dataset_from_context,
        write_only=True,
    )
    set = serializers.SlugRelatedField(queryset=DatasetSet.objects.none(), slug_field="name")

    class Meta:
        model = DatasetElement
        fields = ("set", "element", "element_id", "dataset")
        read_only_fields = ("element",)
        validators = [
            UniqueTogetherValidator(
                queryset=DatasetElement.objects.all(),
                fields=["element_id", "set"],
                message="This element is already part of this set.",
            )
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if dataset := self.context.get("dataset"):
            self.fields["element_id"].queryset = Element.objects.filter(corpus=dataset.corpus)
            self.fields["set"].queryset = dataset.sets.using("default")

    def validate_element_id(self, element):
        dataset = self.context.get("dataset")
        if dataset and dataset.unique_elements and (
            existing_set := (
                dataset.sets
                .filter(set_elements__element=element)
                .values_list("name", flat=True)
                .first()
            )
        ):
            raise ValidationError([
                f"The dataset requires unique elements and this element is already present in set {existing_set}."
            ])
        return element

    def validate(self, data):
        data = super().validate(data)
        data.pop("dataset")
        return data


class DatasetElementInfoSerializer(DatasetElementSerializer):
    dataset = serializers.PrimaryKeyRelatedField(
        queryset=Dataset.objects.none(),
        style={"base_template": "input.html"},
    )
    set = serializers.CharField(max_length=50)
    element = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.none(),
        style={"base_template": "input.html"},
    )

    class Meta(DatasetElementSerializer.Meta):
        fields = DatasetElementSerializer.Meta.fields + ("dataset",)


class ElementDatasetSetSerializer(serializers.ModelSerializer):
    dataset = DatasetSerializer(source="set.dataset")
    previous = serializers.UUIDField(allow_null=True, read_only=True)
    next = serializers.UUIDField(allow_null=True, read_only=True)
    set = serializers.SlugRelatedField(slug_field="name", read_only=True)

    class Meta:
        model = DatasetElement
        fields = ("dataset", "set", "previous", "next")
        read_only_fields = ("previous", "next")


class SelectionDatasetElementSerializer(serializers.Serializer):
    set_id = serializers.PrimaryKeyRelatedField(
        queryset=DatasetSet.objects.none(),
        source="set",
        write_only=True,
        help_text="UUID of a dataset set the elements will be added to.",
        style={"base_template": "input.html"},
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # for openAPI schema generation
        if "request" not in self.context:
            return
        self.fields["set_id"].queryset = DatasetSet.objects.filter(
            dataset__corpus_id__in=Corpus.objects.readable(self.context["request"].user)
        ).select_related("dataset")

    def validate_set_id(self, set):
        if (
            (corpus := self.context.get("corpus"))
            and set.dataset.corpus_id != corpus.id
        ):
            raise ValidationError(f"Dataset {set.dataset.id} is not part of corpus {corpus.name}.")
        if set.dataset.state == DatasetState.Complete:
            raise ValidationError(f"Dataset {set.dataset.id} is marked as completed.")
        # Ensure adding elements to the dataset does not break uniqueness constraint
        selection = self.context["request"].user.selected_elements.filter(corpus=corpus)
        if set.dataset.unique_elements and (
            existing_set := (
                set.dataset.sets
                .exclude(id=set.id)
                .filter(set_elements__element_id__in=selection.values_list("id", flat=True))
                .values_list("name", flat=True)
                .first()
            )
        ):
            raise ValidationError([
                f"The dataset requires unique elements and some elements are already present in set {existing_set}."
            ])
        return set

    def create(self, validated_data):
        user = self.context["request"].user
        corpus = self.context["corpus"]

        DatasetElement.objects.bulk_create(
            (
                DatasetElement(element_id=elt_id, **validated_data)
                for elt_id in (
                    user.selected_elements
                    .filter(corpus=corpus)
                    .values_list("id", flat=True)
                )
            ),
            ignore_conflicts=True
        )
        return validated_data


class DatasetPopulateSerializer(serializers.Serializer):
    parent_id = serializers.UUIDField(
        write_only=True,
        default=None,
        source="parent",
        help_text=dedent("""
            UUID of a parent element from which to select random children. The parent must be in the same corpus as the dataset.

            If this is unset, all elements from the corpus of the dataset will be used.
        """),

    )
    recursive = serializers.BooleanField(
        write_only=True,
        default=False,
        help_text=dedent("""
            When `parent_id` is set, also choose elements among its grandchildren, and so on.

            When `parent_id` is unset, take every element in the corpus instead of only the top-level ones.
        """),
    )
    types = serializers.ListField(
        child=serializers.SlugField(),
        write_only=True,
        allow_empty=False,
        default=["page"],
        help_text="Restrict the selected elements to those with any of the specified element type slugs.",
    )
    count = serializers.IntegerField(
        write_only=True,
        default=1000,
        help_text=dedent("""
            Number of elements to fill the dataset with.

            This cannot exceed the number of available elements.
        """),
        min_value=1,
    )
    sets = serializers.DictField(
        child=serializers.FloatField(min_value=0, max_value=1),
        write_only=True,
        allow_empty=False,
        default={"train": 0.8, "dev": 0.1, "test": 0.1},
        help_text=dedent("""
            A dictionary mapping set names to ratios, defining how the elements will be distributed between the sets.

            The sum of all ratios must be equal to 1.
        """),
    )

    def validate_parent_id(self, parent_id):
        if parent_id is None:
            return
        try:
            parent = self.context["dataset"].corpus.elements.only("id").get(id=parent_id)
        except Element.DoesNotExist:
            raise ValidationError("This element does not exist in the corpus of the dataset.")
        return parent

    def validate_types(self, type_slugs):
        type_slugs = set(type_slugs)
        # Types are prefetched at view's level
        existing_types = {
            t.slug: t.id
            for t in self.context["dataset"].corpus.types.all()
        }
        if (missing_types := [slug for slug in type_slugs if slug not in existing_types]):
            raise ValidationError(f"Some types does not exist in the corpus of the dataset: {sorted(missing_types)}.")
        return [existing_types[slug] for slug in type_slugs]

    def validate_sets(self, sets):
        # Sets are prefetched at view's level
        existing_sets = {
            d.name: d.id
            for d in self.context["dataset"].sets.all()
        }
        errors = []
        if (missing_sets := [s for s in sets if s not in existing_sets]):
            errors.append(f"Some sets does not exist in the dataset: {sorted(missing_sets)}.")
        # Compare values with limited floating point precision
        if not isclose(sum(sets.values()), 1):
            errors.append("The sum of all ratios must be equal to 1.")
        if errors:
            raise ValidationError(errors)
        return {existing_sets[s]: v for s, v in sets.items()}

    def validate(self, data):
        data = super().validate(data)
        elts_count = self.filter_elements(**data).count()
        errors = defaultdict(list)
        if data["count"] > elts_count:
            errors["count"].append(f"This value exceeds the number of filtered elements ({elts_count}).")

        min_ratio = 1 / data["count"]
        if any(ratio < min_ratio for ratio in data["sets"].values()):
            errors["sets"].append(f"Ratio must be greater than {min_ratio:.2g} to contain at least one element.")

        # Ensure none of filtered elements is already part of any
        # set on to avoid any edge case and preserve valid ratio.
        if (
            self.filter_elements(**data)
            .filter(dataset_elements__set__dataset_id=self.context["dataset"].id)
            .exists()
        ):
            errors["__all__"].append("Some filtered elements are already part of a set on this dataset.")

        if errors:
            raise ValidationError(errors)

        return data

    def filter_elements(self, *, parent=None, recursive=False, types, **kwargs):
        qs = self.context["dataset"].corpus.elements.all()
        filters = {}
        if parent:
            if recursive:
                filters["paths__path__overlap"] = [parent.id]
            else:
                # Only pick direct children
                filters["paths__path__last"] = parent.id
        elif not recursive:
            # Only pick top-level elements
            filters["paths__path__last__isnull"] = True
        filters["type_id__in"] = types
        return qs.filter(**filters)


    @transaction.atomic
    def save(self):
        count = self.validated_data["count"]
        sets = self.validated_data["sets"]
        element_ids = list(
            self.filter_elements(**self.validated_data)
            .order_by("?")
            .values_list("id", flat=True)
            [:count]
        )

        dataset_elements = []
        index = 0
        for set_id, ratio in sets.items():
            set_count = int(ratio * count)
            dataset_elements.extend(
                DatasetElement(set_id=set_id, element_id=elt_id)
                for elt_id in element_ids[index:index + set_count]
            )
            index += set_count

        # Distribute remaining elements among sets, as we rounded down using int.
        # Populate largest sets first, to be closer to the ratio.
        sorted_sets = sorted(sets.items(), key=lambda item: -item[1])
        dataset_elements.extend(
            DatasetElement(set_id=set_id, element_id=elt_id)
            for elt_id, (set_id, _ratio) in zip(element_ids[index:], sorted_sets)
        )

        DatasetElement.objects.bulk_create(dataset_elements)
